package org.postgresql;

import java.io.*;
import java.lang.*;
import java.net.*;
import java.util.*;
import java.sql.*;
import org.postgresql.*;
import org.postgresql.core.*;
import org.postgresql.util.*;
import com.sun.net.ssl.*;
import java.security.KeyStore;
import javax.net.ssl.*;
import javax.security.cert.X509Certificate;
/**
 * @version 1.0a 20-DEC-2002 (basic support for SSL communication)
 *               Istvan Nagy <nistvan@ecity.agria.hu>
 * 
 *               This is a prototype of a PG_Stream class that 
 *               supports both normal and SSL connections. 
 *
 * This class is used by Connection & PGlobj for communicating with the
 * backend over normal or a _direct_ SSL connection. You must have JDK
 * with JSSE to compile this driver
 *
 * @see java.sql.Connection
 *
 */
//  This class handles all the Streamed I/O for a org.postgresql connection
public class PG_Stream
{
  private SSLSocket sslconnection;
  private Socket connection;
  // indicator: ssl connection -> true, normal connection -> false  
  private boolean sslconn;
  
  BytePoolDim1 bytePoolDim1 = new BytePoolDim1();
  BytePoolDim2 bytePoolDim2 = new BytePoolDim2();

  private BufferedInputStream pg_input;
  private BufferedOutputStream pg_output;

  // constants' names and values from the original C sources and headers
  private final int NEGOTIATE_SSL_CODE=(1234<<16)|5679;
  private final int SM_DATABASE=64;
  private final int SM_USER=32;
  private final int SM_OPTIONS=64;
  private final int SM_UNUSED=64;
  private final int SM_TTY=64;

  // Converter methods of this class send packets immediately to the stream
  // They work very well so I don't wan to tamper with them. It is a copy
  // of PG_Stream.SendInteger() with an insignificant modification.
  private byte[] create_packetpart(int val,int size) {
    BytePoolDim1 bpd1 = new BytePoolDim1();
    byte b[]=bpd1.allocByte(size);
    while(size-->0) {
	b[size]=(byte)(val&0xff);
	val>>=8;
    }
    return b;
  }
   /**
   * Constructor:  Connect to the PostgreSQL back end and return
   * a stream connection over either a secure or an insecure socket.
   *
   * @param host the hostname to connect to
   * @param port the port number that the postmaster is sitting on
   * @param kspw the password of the keystore
   * @param kspath the path of the keystore
   * @exception IOException if an IOException occurs below it.
   *
   * See the $PGSQL_SRC_ROOT/backend/libpq/README.SSL and the
   * $PGSQL_SRCROOT/interfaces/libpq/fe-connect.c to study how
   * the ssl code negotiation happens from a client's perspective
   */
  public PG_Stream(String host,int port, String kspw, String kspath) throws IOException {
    try {
	if ( kspw.equals("") || kspath.equals("") ) {
	    // normal connection (the original code)
	    sslconn=false;
	    
	    // establish a normal connection to the back end
	    connection = new Socket(host, port);
	    
	    // Submitted by Jason Venner <jason@idiom.com> adds a 10x speed
	    // improvement on FreeBSD machines (caused by a bug in their TCP Stack)
	    connection.setTcpNoDelay(true);
	    
	    // Buffer sizes submitted by Sverre H Huseby <sverrehu@online.no>
	    pg_input = new BufferedInputStream(connection.getInputStream(), 8192);
	    pg_output = new BufferedOutputStream(connection.getOutputStream(), 8192);
	
	} else {
	    // try to establish an ssl connection to the back end
	    Socket s = new Socket(host,port);
	    
	    // buffer sizes submitted by Sverre H Huseby <sverrehu@online.no>
	    BufferedOutputStream bos = new BufferedOutputStream(s.getOutputStream(),8192);
	    BufferedInputStream bis = new BufferedInputStream(s.getInputStream(),8192);

	    // start handshake with the PostgreSQL server and decide on the basis
	    // of its answer wheter the back end supports SSL or not.
	
	    // send the size of the startup packet (assembled from the size of its parts) 
	    // - see the chapter 4.4 (Message Formats) in PostgreSQL's documentation 
	    // The first '4' is the size of this message, the second '4' is the size of 
	    // the 'NEGOTIATE_SSL_CODE' messagepart (in bytes).  
	    bos.write(create_packetpart(4+4+SM_DATABASE+SM_USER+SM_OPTIONS+SM_UNUSED+SM_TTY,4));

	    // I send NEGOTIATE_SLL_CODE despite of PG_PROTOCOL_LATEST_MAJOR 
	    // and PG_PROTOCOL_LATEST_MINOR codes
	    bos.write(create_packetpart(NEGOTIATE_SSL_CODE,4));

	    // send zeros in the rest of the startup packet. Authentication 
	    // is carried out by Connection.openConnection()
	    for(int ii=0;ii<SM_DATABASE+SM_USER+SM_OPTIONS+SM_UNUSED+SM_TTY;++ii)
		bos.write(0);
	    bos.flush();

	    // backend's answer
	    char ssl_answer=(char)bis.read();
	
	    // if the answer is 'S' then SSL is supported by PostgreSQL
	    if(ssl_answer=='S'){
		/*
	    	    I don't use the default factories because I will use my own keystore
	    	    instead of the JVM's default keystore. By this way, I can use
		    any certificate - for example my self-signed certificates -
	    	    that I have marked as a trusted certificate in my own keystore.
	    	    Untrusted certs cause an Exception in JSSE. Sun's JSSE considers
	    	    such an Exception as a fatal problem and closes the socket.
	    	    Other JSSE implementations may behave in a different way(?).
	    	    See JSSE ApiDoc for details (part of the JSSE distribution)
		*/
		sslconn=true;
		
		// PostgreSQL supports SSLv3
		SSLContext ssl_ctx=SSLContext.getInstance("TLS");

		// We use a TrustManager that supports X.509
		TrustManagerFactory tmf=TrustManagerFactory.getInstance("SunX509");

		// my keystore use the jks format
		// This param may be taken from a property to support different
		// keystore formats (why?). 
		KeyStore ks=KeyStore.getInstance("JKS");

		// load the keystore
		char[] keystore_pwd=kspw.toCharArray();
		ks.load(new FileInputStream(kspath),keystore_pwd);

		// init the TrustManagerFactory instance
		tmf.init(ks);

		// init the SLContext instance 
		ssl_ctx.init(null,tmf.getTrustManagers(),null);

		// get the SocketFactory instance that supports my own settings
		SSLSocketFactory sf=(SSLSocketFactory)ssl_ctx.getSocketFactory();

		// create a secure socket
		sslconnection=(SSLSocket)sf.createSocket(s,host,port,true);

		// start the SSL handshaking process
		sslconnection.startHandshake();
		
		// submitted by Jason Venner <jason@idiom.com>
    		sslconnection.setTcpNoDelay(true);
	
		// buffer size submitted by Sverre H Huseby <sverrehu@online.no> 
		pg_input = new BufferedInputStream(sslconnection.getInputStream(), 8192);
		pg_output = new BufferedOutputStream(sslconnection.getOutputStream(), 8192);
	    
	    } else {
		/* 
		    to my mind, if PostgreSQL refuses an SSL connection 
		    then I can try to continue with a normal connection(answ.:'E' or 'N'). 
		    If backend didn't listen to connections(to the 'port' port, 
		    or it didn't operate at all) then an Exception will be thrown,
		    unless other daemon was listening to the 'port' port answering 
		    my message with a valid character ('S' or 'E' or 'N'). It is unlikely.
		*/
		sslconn=false;
		connection=s;
		// submitted by Jason Venner <jason@idiom.com>
    		connection.setTcpNoDelay(true);
		// Buffer sizes submitted by Sverre H Huseby <sverrehu@online.no>
		pg_input = new BufferedInputStream(connection.getInputStream(), 8192);
		pg_output = new BufferedOutputStream(connection.getOutputStream(), 8192);
	    }
	}
    } catch (Exception e) {
	// should a problem occures it will be thrown as an IOException
	throw new IOException(e.getMessage());
    }
  }
  
  /**
   * Sends a single character to the back end
   *
   * @param val the character to be sent
   * @exception IOException if an I/O error occurs
   */
  public void SendChar(int val) throws IOException
  {
      // Original code
      //byte b[] = new byte[1];
      //b[0] = (byte)val;
      //pg_output.write(b);

      // Optimised version by Sverre H. Huseby Aug 22 1999 Applied Sep 13 1999
      pg_output.write((byte)val);
  }

  /**
   * Sends an integer to the back end
   *
   * @param val the integer to be sent
   * @param siz the length of the integer in bytes (size of structure)
   * @exception IOException if an I/O error occurs
   */
  public void SendInteger(int val, int siz) throws IOException
  {
    byte[] buf = bytePoolDim1.allocByte(siz);

    while (siz-- > 0)
      {
	buf[siz] = (byte)(val & 0xff);
	val >>= 8;
      }
    Send(buf);
  }

  /**
   * Sends an integer to the back end in reverse order.
   *
   * This is required when the backend uses the routines in the
   * src/backend/libpq/pqcomprim.c module.
   *
   * As time goes by, this should become obsolete.
   *
   * @param val the integer to be sent
   * @param siz the length of the integer in bytes (size of structure)
   * @exception IOException if an I/O error occurs
   */
  public void SendIntegerReverse(int val, int siz) throws IOException
  {
    byte[] buf = bytePoolDim1.allocByte(siz);
    int p=0;
    while (siz-- > 0)
      {
	buf[p++] = (byte)(val & 0xff);
	val >>= 8;
      }
    Send(buf);
  }

  /**
   * Send an array of bytes to the backend
   *
   * @param buf The array of bytes to be sent
   * @exception IOException if an I/O error occurs
   */
  public void Send(byte buf[]) throws IOException
  {
    pg_output.write(buf);
  }

  /**
   * Send an exact array of bytes to the backend - if the length
   * has not been reached, send nulls until it has.
   *
   * @param buf the array of bytes to be sent
   * @param siz the number of bytes to be sent
   * @exception IOException if an I/O error occurs
   */
  public void Send(byte buf[], int siz) throws IOException
  {
    Send(buf,0,siz);
  }

  /**
   * Send an exact array of bytes to the backend - if the length
   * has not been reached, send nulls until it has.
   *
   * @param buf the array of bytes to be sent
   * @param off offset in the array to start sending from
   * @param siz the number of bytes to be sent
   * @exception IOException if an I/O error occurs
   */
  public void Send(byte buf[], int off, int siz) throws IOException
  {
    int i;

    pg_output.write(buf, off, ((buf.length-off) < siz ? (buf.length-off) : siz));
    if((buf.length-off) < siz)
      {
	for (i = buf.length-off ; i < siz ; ++i)
	  {
	    pg_output.write(0);
	  }
      }
  }

  /**
   * Sends a packet, prefixed with the packet's length
   * @param buf buffer to send
   * @exception SQLException if an I/O Error returns
   */
  public void SendPacket(byte[] buf) throws IOException
  {
    SendInteger(buf.length+4,4);
    Send(buf);
  }

  /**
   * Receives a single character from the backend
   *
   * @return the character received
   * @exception SQLException if an I/O Error returns
   */
  public int ReceiveChar() throws SQLException
  {
    int c = 0;

    try
      {
	c = pg_input.read();
	if (c < 0) throw new PSQLException("postgresql.stream.eof");
      } catch (IOException e) {
	throw new PSQLException("postgresql.stream.ioerror",e);
      }
      return c;
  }

  /**
   * Receives an integer from the backend
   *
   * @param siz length of the integer in bytes
   * @return the integer received from the backend
   * @exception SQLException if an I/O error occurs
   */
  public int ReceiveInteger(int siz) throws SQLException
  {
    int n = 0;

    try
      {
	for (int i = 0 ; i < siz ; i++)
	  {
	    int b = pg_input.read();

	    if (b < 0)
	      throw new PSQLException("postgresql.stream.eof");
	    n = n | (b << (8 * i)) ;
	  }
      } catch (IOException e) {
	throw new PSQLException("postgresql.stream.ioerror",e);
      }
      return n;
  }

  /**
   * Receives an integer from the backend
   *
   * @param siz length of the integer in bytes
   * @return the integer received from the backend
   * @exception SQLException if an I/O error occurs
   */
  public int ReceiveIntegerR(int siz) throws SQLException
  {
    int n = 0;

    try
      {
	for (int i = 0 ; i < siz ; i++)
	  {
	    int b = pg_input.read();

	    if (b < 0)
	      throw new PSQLException("postgresql.stream.eof");
	    n = b | (n << 8);
	  }
      } catch (IOException e) {
	throw new PSQLException("postgresql.stream.ioerror",e);
      }
      return n;
  }

  /**
   * Receives a null-terminated string from the backend.  Maximum of
   * maxsiz bytes - if we don't see a null, then we assume something
   * has gone wrong.
   *
   * @param maxsiz maximum length of string
   * @return string from back end
   * @exception SQLException if an I/O error occurs
   */
  public String ReceiveString(int maxsiz) throws SQLException
  {
    byte[] rst = bytePoolDim1.allocByte(maxsiz);
    return ReceiveString(rst, maxsiz, null);
  }

  /**
   * Receives a null-terminated string from the backend.  Maximum of
   * maxsiz bytes - if we don't see a null, then we assume something
   * has gone wrong.
   *
   * @param maxsiz maximum length of string
   * @param encoding the charset encoding to use.
   * @param maxsiz maximum length of string in bytes
   * @return string from back end
   * @exception SQLException if an I/O error occurs
   */
  public String ReceiveString(int maxsiz, String encoding) throws SQLException
  {
    byte[] rst = bytePoolDim1.allocByte(maxsiz);
    return ReceiveString(rst, maxsiz, encoding);
  }

  /**
   * Receives a null-terminated string from the backend.  Maximum of
   * maxsiz bytes - if we don't see a null, then we assume something
   * has gone wrong.
   *
   * @param rst byte array to read the String into. rst.length must
   *        equal to or greater than maxsize.
   * @param maxsiz maximum length of string in bytes
   * @param encoding the charset encoding to use.
   * @return string from back end
   * @exception SQLException if an I/O error occurs
   */
  public String ReceiveString(byte rst[], int maxsiz, String encoding)
      throws SQLException
  {
    int s = 0;

    try
      {
	while (s < maxsiz)
	  {
	    int c = pg_input.read();
	    if (c < 0)
	      throw new PSQLException("postgresql.stream.eof");
 	    else if (c == 0) {
 		rst[s] = 0;
 		break;
 	    } else
	      rst[s++] = (byte)c;
	  }
	if (s >= maxsiz)
	  throw new PSQLException("postgresql.stream.toomuch");
      } catch (IOException e) {
	throw new PSQLException("postgresql.stream.ioerror",e);
      }
      String v = null;
      if (encoding == null)
          v = new String(rst, 0, s);
      else {
          try {
              v = new String(rst, 0, s, encoding);
          } catch (UnsupportedEncodingException unse) {
              throw new PSQLException("postgresql.stream.encoding", unse);
          }
      }
      return v;
  }

  /**
   * Read a tuple from the back end.  A tuple is a two dimensional
   * array of bytes
   *
   * @param nf the number of fields expected
   * @param bin true if the tuple is a binary tuple
   * @return null if the current response has no more tuples, otherwise
   *	an array of strings
   * @exception SQLException if a data I/O error occurs
   */
  public byte[][] ReceiveTuple(int nf, boolean bin) throws SQLException
  {
    int i, bim = (nf + 7)/8;
    byte[] bitmask = Receive(bim);
    byte[][] answer = bytePoolDim2.allocByte(nf);

    int whichbit = 0x80;
    int whichbyte = 0;

    for (i = 0 ; i < nf ; ++i)
      {
	boolean isNull = ((bitmask[whichbyte] & whichbit) == 0);
	whichbit >>= 1;
	if (whichbit == 0)
	  {
	    ++whichbyte;
	    whichbit = 0x80;
	  }
	if (isNull)
	  answer[i] = null;
	else
	  {
	    int len = ReceiveIntegerR(4);
	    if (!bin)
	      len -= 4;
	    if (len < 0)
	      len = 0;
	    answer[i] = Receive(len);
	  }
      }
    return answer;
  }

  /**
   * Reads in a given number of bytes from the backend
   *
   * @param siz number of bytes to read
   * @return array of bytes received
   * @exception SQLException if a data I/O error occurs
   */
  private byte[] Receive(int siz) throws SQLException
  {
      byte[] answer = bytePoolDim1.allocByte(siz);
    Receive(answer,0,siz);
    return answer;
  }

  /**
   * Reads in a given number of bytes from the backend
   *
   * @param buf buffer to store result
   * @param off offset in buffer
   * @param siz number of bytes to read
   * @exception SQLException if a data I/O error occurs
   */
  public void Receive(byte[] b,int off,int siz) throws SQLException
  {
    int s = 0;

    try
      {
	while (s < siz)
	  {
	    int w = pg_input.read(b, off+s, siz - s);
	    if (w < 0)
	      throw new PSQLException("postgresql.stream.eof");
	    s += w;
	  }
      } catch (IOException e) {
	  throw new PSQLException("postgresql.stream.ioerror",e);
      }
  }

  /**
   * This flushes any pending output to the backend. It is used primarily
   * by the Fastpath code.
   * @exception SQLException if an I/O error occurs
   */
  public void flush() throws SQLException
  {
    try {
      pg_output.flush();
    } catch (IOException e) {
      throw new PSQLException("postgresql.stream.flush",e);
    }
  }

  /**
   * Closes the connection
   *
   * @exception IOException if a IO Error occurs
   */
  public void close() throws IOException
  {
    pg_output.write("X\0".getBytes());
    pg_output.flush();
    pg_output.close();
    pg_input.close();
    // close an ssl or a normal socket depending on the value of the mode indicator
    if (sslconn) 
	sslconnection.close(); 
    else
	connection.close();
  }

}

